import datetime
import logging
import os

import google.cloud.storage as storage
import jwt
import requests
from gcsfs.retry import HttpError

from dapla.auth import AuthClient

logger = logging.getLogger(__name__)


class Doctor:
    """Class of functions that perform checks on Dapla.

    Checks whether user is authenticated, if the keycloak-token is valid and if user has access to GCS.
    Each method can be run individually or collectively with the 'health' method.
    """

    @staticmethod
    def jupyterhub_auth_valid() -> bool:
        """Checks wheter user is logged in and authenticated to Jupyterhub."""
        print("Checking authentication to JupyterHub...")
        try:
            # Attempt fetching the Jupyterhub user
            AuthClient.fetch_local_user_from_jupyter()
        except Exception:
            return False
        return True

    @staticmethod
    def keycloak_token_valid() -> bool:
        """Checks whether the keycloak token is valid by attempting to access a keycloak-token protected service."""
        print("Checking that your keycloak token is valid...")
        algorithms = ["RS256"]

        keycloak_token = AuthClient.fetch_personal_token()

        claims = jwt.decode(
            keycloak_token, algorithms=algorithms, options={"verify_signature": False}
        )

        if not Doctor._is_token_expired(claims):
            return True
        else:
            return False

    @staticmethod
    def _is_token_expired(token: dict[str, str]) -> bool:
        try:
            # get the expiry time from the payload
            expiry_time = token["exp"]

            # convert the expiry time to a datetime object
            expiry_timestamp = datetime.datetime.fromtimestamp(float(expiry_time))

            # if the current time is greater than the expiry time, the token is expired
            return datetime.datetime.now() > expiry_timestamp
        except KeyError:
            # if the token does not have an expiry time (exp claim), consider it not expired
            return False

    @staticmethod
    def gcs_credentials_valid() -> bool:
        """Checks whether the users google cloud storage token is valid by accessing a GCS service."""
        print("Checking your Google Cloud Storage credentials...")

        # Fetch the google token
        google_token, _ = AuthClient.fetch_google_credentials().token

        try:
            requests.get(
                "https://oauth2.googleapis.com/tokeninfo?access_token=%s" % google_token
            )
        except HttpError as ex:
            if str(ex) == "Invalid Credentials, 401":
                return False
            else:
                logger.exception(ex)

        return True

    @staticmethod
    def bucket_access() -> bool:
        """Checks whether user has access to a common google bucket."""
        print(
            "Checking that you have access to a common Google Cloud Storage bucket..."
        )

        # Fetch google credentials and create client object
        client = storage.Client(credentials=AuthClient.fetch_google_credentials())

        # Set the bucket that is to be accessed
        if os.environ["CLUSTER_ID"] == "staging-bip-app":
            bucket = "ssb-staging-dapla-felles-data-delt"
        else:
            bucket = "ssb-prod-dapla-felles-data-delt"

        try:
            # Attempt to access the bucket
            client.get_bucket(bucket)
        except Exception:
            return False
        else:
            return True

    @classmethod
    def health(cls) -> None:
        """Runs a series of checks to determine the health of Dapla setup."""
        print("Performing checks...")
        if not cls.jupyterhub_auth_valid():
            print("You are either not logged in or not authenticated to JupyterHub.")
            exit(1)
        if not cls.keycloak_token_valid():
            print("Your keycloak token seems to be expired.")
            exit(1)
        if not cls.gcs_credentials_valid():
            print(
                "Your GCS credentials seem to be invalid. Dapla Doctor could not fetch GCS bucket."
            )
            exit(1)
        if not cls.bucket_access():
            print("You do not seem to have access to a common GCS bucket.")
            exit(1)
        print("Everything seems to be in order.")



import typing as t
from typing import Any
from typing import Optional

import gcsfs
from google.oauth2.credentials import Credentials


class GCSFileSystem(gcsfs.GCSFileSystem):  # type: ignore [misc]
    """GCSFileSystem is a wrapper around gcsfs.GCSFileSystem."""

    def __init__(
        self, token: Optional[dict[str, str] | str | Credentials] = None, **kwargs: Any
    ) -> None:
        """Initialize GCSFileSystem."""
        super().__init__(token=token, **kwargs)
        # Temporary bug fix for https://issues.apache.org/jira/browse/ARROW-7867
        # Spark writes an empty file to GCS (to mimic a folder structure) before writing partitioned data
        # Resolve this by ignoring the "empty" file when reading partitioned parquet files
        try:
            # Constant is moved to core module in Pyarrow 10.0.0
            from pyarrow.parquet.core import (  # type: ignore [attr-defined]
                EXCLUDED_PARQUET_PATHS,
            )
        except ImportError:
            # Fallback for Pyarrow versions <10.0.0
            from pyarrow.parquet.core import (  # type: ignore [attr-defined]
                EXCLUDED_PARQUET_PATHS,
            )
        from pyarrow.parquet import ParquetManifest

        EXCLUDED_PARQUET_PATHS.add("")
        ParquetManifest._should_silently_exclude = (  # type: ignore [attr-defined]
            GCSFileSystem._should_silently_exclude
        )

    def isdir(self, path: str) -> bool:
        """Check if path is a directory."""
        info = super(gcsfs.GCSFileSystem, self).info(path)
        return t.cast(bool, info["type"] == "directory")

    @staticmethod
    # This code is from from pyarrow.parquet.core
    def _should_silently_exclude(file_name: str) -> bool:
        try:
            # Constant is moved to core module in Pyarrow 10.0.0
            from pyarrow.parquet.core import (  # type: ignore [attr-defined]
                EXCLUDED_PARQUET_PATHS,
            )
        except ImportError:
            # Fallback for Pyarrow versions <10.0.0
            from pyarrow.parquet.core import (  # type: ignore [attr-defined]
                EXCLUDED_PARQUET_PATHS,
            )

        return (
            file_name.endswith(".crc")
            or file_name.endswith("_$folder$")  # Checksums
            or file_name.startswith(".")  # HDFS directories in S3
            or file_name.startswith("_")  # Hidden files starting with .
            or ".tmp" in file_name  # Hidden files starting with _
            or file_name in EXCLUDED_PARQUET_PATHS  # Temp files
        )



import json
import os
import typing as t
from collections.abc import Sequence
from datetime import datetime
from datetime import timedelta
from functools import lru_cache
from typing import Any
from typing import Optional

import google.auth
import requests
from google.auth.transport.requests import Request as GoogleAuthRequest
from google.oauth2.credentials import Credentials
from IPython.display import HTML
from IPython.display import display
from jupyterhub.services.auth import HubAuth

# Refresh window was modified in: https://github.com/googleapis/google-auth-library-python/commit/c6af1d692b43833baca978948376739547cf685a
# The change was directed towards high latency environments, and should not apply to us.
# Since we can't force a refresh, the threshold is lowered to keep us from waiting ~4 minutes for a new token.
# A permanent fix would be to supply credentials with a refresh endpoint
# that allways returns a token that is valid for more than 3m 45s.
google.auth._helpers.REFRESH_THRESHOLD = timedelta(seconds=20)


class AuthClient:
    """Client for retrieving authentication information."""

    @staticmethod
    def fetch_google_token_from_oidc_exchange(
        request: GoogleAuthRequest, _scopes: Sequence[str]
    ) -> tuple[str, datetime]:
        """Fetches the Google token by exchanging an OIDC token.

        Args:
            request: The GoogleAuthRequest object.
            _scopes: The scopes to request.

        Raises:
            AuthError: If the request to the OIDC token exchange endpoint fails.

        Returns:
            A tuple of (google-token, expiry).
        """
        response = request.__call__(
            url=os.environ["OIDC_TOKEN_EXCHANGE_URL"],
            method="POST",
            body={
                "subject_token": os.environ["OIDC_TOKEN"],
                "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
                "requested_issuer": "google",
                "client_id": "onyxia",
            },
        )
        if response.status == 200:
            auth_data = json.loads(response.data)
            expiry = datetime.utcnow() + timedelta(seconds=auth_data["expires_in"])
            return auth_data["access_token"], expiry
        else:
            raise AuthError

    @staticmethod
    def fetch_google_token_from_jupyter() -> str:
        """Fetches the personal access token for the current user.

        Raises:
            AuthError: If the token exchange request to JupyterHub fails.

        Returns:
            The personal access token.
        """
        try:
            google_token = AuthClient.fetch_local_user_from_jupyter()[
                "exchanged_tokens"
            ]["google"]["access_token"]
            return t.cast(str, google_token)
        except AuthError as err:
            err._print_warning()
            raise err

    @staticmethod
    def fetch_local_user_from_jupyter() -> dict[str, Any]:
        """Retrieves user information, most notably access tokens for use in authentication.

        Raises:
            AuthError: If the request to the user endpoint fails.

        Returns:
            The user data from the token .
        """
        # Helps getting the correct ssl configs
        hub = HubAuth()
        response = requests.get(
            os.environ["LOCAL_USER_PATH"],
            headers={"Authorization": "token %s" % hub.api_token},
            cert=(hub.certfile, hub.keyfile),
            verify=hub.client_ca,
            allow_redirects=False,
        )
        if response.status_code == 200:
            user_data = response.json()
            return t.cast(dict[str, Any], user_data)
        else:
            raise AuthError

    @staticmethod
    def fetch_google_credentials() -> Credentials:
        """Fetches the Google credentials for the current user.

        Returns:
            The Google "Credentials" object.
        """
        if AuthClient.is_ready():
            try:

                def _refresh_handler(
                    request: google.auth.transport.Request, scopes: Sequence[str]
                ) -> tuple[str, datetime]:
                    # We manually override the refresh_handler method with our custom logic for fetching tokens.
                    # Previously, we directly overrode the `refresh` method. However, this
                    # approach led to deadlock issues in gcsfs/credentials.py's maybe_refresh method.
                    return AuthClient.fetch_google_token()

                token, expiry = AuthClient.fetch_google_token()
                credentials = Credentials(
                    token=token,
                    expiry=expiry,
                    token_uri="https://oauth2.googleapis.com/token",
                    refresh_handler=_refresh_handler,
                )
            except AuthError as err:
                err._print_warning()

            return credentials
        else:
            # Fetch credentials from Google Cloud SDK
            credentials, _ = google.auth.default()
            return credentials

    @staticmethod
    def fetch_personal_token() -> str:
        """Fetches the personal access token for the current user."""
        try:
            personal_token = AuthClient.fetch_local_user_from_jupyter()["access_token"]
            return t.cast(str, personal_token)
        except AuthError as err:
            err._print_warning()
            raise err

    @staticmethod
    @lru_cache(maxsize=1)
    def fetch_email_from_credentials() -> Optional[str]:
        """Retrieves an e-mail based on current Google Credentials. Potentially makes a Google API call."""
        if AuthClient.is_ready():
            credentials = AuthClient.fetch_google_credentials()
            response = requests.get(
                url=f"https://oauth2.googleapis.com/tokeninfo?access_token={credentials.token}"
            )

            return response.json().get("email") if response.status_code == 200 else None

        else:
            user_info = AuthClient.fetch_local_user_from_jupyter()
            return user_info.get("username")

    @staticmethod
    def fetch_google_token(
        request: Optional[GoogleAuthRequest] = None,
        scopes: Optional[Sequence[str]] = None,
    ) -> tuple[str, datetime]:
        """Fetches the Google token for the current user.

        Scopes in the argument is ignored, but are kept for compatibility
        with the Credentials refresh handler method signature.

        Args:
            request: The GoogleAuthRequest object.
            scopes: The scopes to request.

        Raises:
            AuthError: If the token exchange request to JupyterHub fails.

        Returns:
            The Google token.
        """
        try:
            if AuthClient._is_oidc_token():
                if request is None:
                    request = GoogleAuthRequest()
                if scopes is None:
                    scopes = []

                google_token, expiry = AuthClient.fetch_google_token_from_oidc_exchange(
                    request, scopes
                )
            else:
                user_info = AuthClient.fetch_local_user_from_jupyter()
                google_token = user_info["exchanged_tokens"]["google"]["access_token"]
                expiry = datetime.utcfromtimestamp(
                    user_info["exchanged_tokens"]["google"]["exp"]
                )

        except AuthError as err:
            err._print_warning()
            raise err

        return google_token, expiry

    @staticmethod
    def is_ready() -> bool:
        """Checks whether the authentication handler can be used."""
        return "LOCAL_USER_PATH" in os.environ or "OIDC_TOKEN" in os.environ

    @staticmethod
    def _is_oidc_token() -> bool:
        return "OIDC_TOKEN" in os.environ


class AuthError(Exception):
    """This exception class is used when the communication with the custom auth handler fails.

    This is normally due to stale auth session.
    """

    def _print_warning(self) -> None:
        (
            display(
                HTML(
                    'Your session has timed out. Please <a href="/hub/login">log in</a> to continue.'
                )  # type: ignore [no-untyped-call]
            )
        )



import json
from typing import Any

import requests
from requests import Response

from .auth import AuthClient


class CollectorClient:
    """Client for working with DataCollector."""

    def __init__(self, collector_url: str) -> None:
        """Initialize CollectorClient."""
        self.collector_url = collector_url

    def start(self, specification: dict[str, Any]) -> Response:
        """Start a new collector task.

        Args:
            specification: The JSON object of the collector specification

        Returns:
            Response: The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()
        collector_response = requests.put(
            self.collector_url,
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
            data=json.dumps(specification),
        )
        collector_response.raise_for_status()
        print(
            "Task initiated successfully! Check running tasks with the running_tasks() method."
        )
        return collector_response

    def running_tasks(self) -> Response:
        """Get all running collector tasks."""
        keycloak_token = AuthClient.fetch_personal_token()
        collector_response = requests.get(
            self.collector_url, headers={"Authorization": "Bearer %s" % keycloak_token}
        )
        collector_response.raise_for_status()
        return collector_response

    def stop(self, task_id: int) -> Response:
        """Stop a running collector task.

        Args:
            task_id: The id of the task to stop.

        Returns:
            Response: The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()
        collector_response = requests.delete(
            f"{self.collector_url}/{task_id}",
            headers={"Authorization": "Bearer %s" % keycloak_token},
        )
        if collector_response.status_code == 400:
            print(
                "Collector task with id: "
                + str(task_id)
                + " cannot be stopped! Maybe it was already stopped or completed"
            )
        else:
            print("Collector task with id: " + str(task_id) + " stopped successfully")

        return collector_response



import json
from typing import Any

import requests
from requests import Response

from .auth import AuthClient


class ConverterClient:
    """Client for working with DataConverter."""

    def __init__(self, converter_url: str) -> None:
        """Initialize ConverterClient."""
        self.converter_url = converter_url

    def start(self, job_config: dict[str, Any]) -> Response:
        """Schedule a new converter job.

        Args:
            job_config: The JSON object of the job configuration

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()
        converter_response = requests.post(
            f"{self.converter_url}/jobs",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
            data=json.dumps(job_config),
        )
        converter_response.raise_for_status()
        return converter_response

    def start_simulation(self, job_config: dict[str, Any]) -> Response:
        """Start a simulated converter job.

        Useful for testing job configurations or diagnosing pseudonymization issues.

        Args:
            job_config: The JSON object of the job configuration

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()
        converter_response = requests.post(
            f"{self.converter_url}/jobs/simulation",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
            data=json.dumps(job_config),
        )
        converter_response.raise_for_status()
        return converter_response

    def get_job_summary(self, job_id: str) -> Response:
        """Retrieve the execution summary for a specific converter job.

        Args:
            job_id: The ID of the job

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()

        job_summary = requests.get(
            f"{self.converter_url}/jobs/{job_id}/execution-summary",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
        )
        job_summary.raise_for_status()
        return job_summary

    def stop_job(self, job_id: str) -> Response:
        """Stop a specific converter job.

        Args:
            job_id: The ID of the job

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()

        job_status = requests.post(
            f"{self.converter_url}/jobs/{job_id}/stop",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
        )
        job_status.raise_for_status()
        return job_status

    def get_pseudo_report(self, job_id: str) -> Response:
        """Get a report with details about how pseudonymization is being applied for a specific job.

        The report includes:
        * mathced pseudo rules
        * unmatched pseudo rules
        * pseudo rule to field match map - rule names and their corresponding matches
        * field matches to pseudo rule map - field names and the corresponding pseudo rule that covers this field
        * metrics
        * textual schema hierarchy that illustrates how the pseudo rules are being applied

        Args:
            job_id: The ID of the job

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()

        pseudo_report = requests.get(
            f"{self.converter_url}/jobs/{job_id}/reports/pseudo",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
        )
        pseudo_report.raise_for_status()
        return pseudo_report

    def get_pseudo_schema(self, job_id: str) -> Response:
        """Get hierarchical schema representation that details how pseudo rules are being applied.

        This is a smaller version get_pseudo_report

        Args:
            job_id: The ID of the job

        Returns:
            The "requests.Response" object from the API call
        """
        keycloak_token = AuthClient.fetch_personal_token()

        pseudo_report = requests.get(
            f"{self.converter_url}/jobs/{job_id}/reports/pseudo-schema-hierarchy",
            headers={
                "Authorization": "Bearer %s" % keycloak_token,
                "Content-type": "application/json",
            },
        )
        pseudo_report.raise_for_status()
        return pseudo_report



import os
from typing import Any
from urllib.parse import urlparse


def uiWebUrl(self: Any) -> str:
    """Fix the Spark UI link to point to the jupyter-server-proxy.

    From https://github.com/jupyterhub/jupyter-server-proxy/issues/57
    """
    web_url = self._jsc.sc().uiWebUrl().get()
    port = urlparse(web_url).port
    return os.environ["JUPYTERHUB_SERVICE_PREFIX"] + f"proxy/{port}/jobs/"
